# Demonstration of code and analysis

# PREAMBLE ===========================================================
rm(list = ls())

# Source function files and set working directory
cat("Please choose the functions.R file in the popup window. Mind that all other files should be in the same folder.")
functions_file <- file.choose() # Choose functions.R file
source(functions_file)
setwd(dirname(functions_file))

# Install and load necessary packages
load_install_package("deSolve")
load_install_package("plyr")
load_install_package("ggplot2")

# CODE ===============================================================

# 1. ANALYSES IN CONSTANT ENVIRONMENT --------------------------------
# 1.1 Simulate a single trajectory

  # set the random number generator seed and
  # trajectory parameters
set.seed(9)
d <- 50       # number of traits
tmax <- 500   # maximum simulation time

  # generate trajectory (ca. 2 minutes)
traj_const_env <- gen_traj(d = d, tmax = tmax)

plot(traj_const_env$time, traj_const_env$traj[,1], type = "l",
     xlab = "Time",
     ylab = expression(x[1]))


# 1.2 Calculate Largest Lyapunov Exponent (LLE)
  # calculate LLE (ca. 12 minutes)
lle <- calc_lle(traj_const_env)

plot(lle$time, lle$lle, type = "l",
     xlab = "Time",
     ylab = expression(lambda))


# 1.3 Calculate local LLE (window average of LLE)
local_lle <- calc_local_lle(lle, 
                            average_window = 200, 
                            burnout = 50)

plot(local_lle$time, local_lle$local_lle, type = "l",
     xlab = "Time",
     ylab = expression(bar(lambda)))


# 1.4 Classify trajectory along time
class_traj <- class_lle(local_lle, criterion = 0.02)

  # visualize result in a data.frame format
class_traj_dat <- data.frame(time = class_traj$time,
                             class = class_traj$class)
View(class_traj_dat)


# 1.5 Fit of exponential models and estimation of asymptotic frequency of chaos

  # import data of frequency of chaos along time for several dimensionalities d
freq_chaos_dat <- read.csv("dat_freq_chaos_const_env.csv")

  # fit exponential model to the data of frequency of chaos along time for each dimensionality d
  # WARNING: errors will be returned for those  dimensionalities for which a fit was not possible
models <- dlply(freq_chaos_dat, "d", function(x){
  out <- NULL
  try(
    out <- with(x, {
      nls(I(freq_chaos)~B*exp(-(1/A)*I(time-400)) + C, 
          start = c(A = 10, B = 0.5, C = 0))
    })
  )
  return(out)
})

# 1.6 Figures of exponential fits
  # vectors required for plotting
time <- sort(unique(freq_chaos_dat$time))
ds <- as.numeric(as.character(sort(unique(freq_chaos_dat$d))))
  
  # make a plot for each dimensionality d
for(i in seq_along(models)){
  cat("\rplot ", i, "of", length(ds))
  cur_dat <- freq_chaos_dat[freq_chaos_dat$d == ds[i],]
  
  p <- ggplot(cur_dat, aes(time, freq_chaos)) +
    geom_line() +
    labs(x = "Time", 
         y = "Frequency of chaos",
         title = sprintf("d = %s", ds[i])) +
    my_theme()
    
  if(is.null(models[[i]])){
    plot(p)
  } else {
    model_dat <- data.frame(time = time,
                            model = predict(models[[i]]))
    p <- p +
      geom_line(data = model_dat, aes(time, model),
                color = "indianred3") +
      geom_hline(yintercept = coef(models[[i]])["C"],
                 lty = 2,
                 color = "indianred3")
    plot(p)
  }
}

# 2. ANALYSES IN CHANGING ENVIRONMENT ---------------------------------
d <- 50 
amplitude <- 5.99
period <- 5

# 2.1 Simulate a single trajectory for a given condition of amplitude and period of optimum oscilation (ca. 0.5 minute)
traj_chang_env <- 
  gen_traj(d = d, tmax = 100, rescale = T,
           direction = rep(1/sqrt(d), d),
           ch.opt.FUN = function(t) ch.opt.sinusoid(t, amplitude,period))

  # extract optimum for plotting
optimum <- get.opt(traj_chang_env)

  # plot trajectory and optimum
ggplot(data.frame(time = traj_chang_env$time,
                  x1 = traj_chang_env$traj[,1],
                  theta1 = optimum[,1]),
       aes(time)) +
  geom_line(aes(y = theta1, color = "theta1")) +
  geom_line(aes(y = x1, color = "x1")) +
  labs(x = "Time", y = expression(paste(x[1],"/",theta[1])),
       title = "Trajectory",
       color = "") +
  scale_color_manual(values = c(x1 = "black", theta1 = "#E89143"),
                     labels = c(x1 = expression(x[1]), theta1 = expression(theta[1]))) +
  my_theme()

# 2.2 Correct the lag and fit the regression model
  
  # find the lag between the phenotype and the optimum time-series by maximizing the R^2 of the regression 
lag <- best_shift(traj_chang_env, period = period)

  # get regression model with the shifted optimum and print results on the console
model <- regression_model(traj_chang_env, shift = lag)

regression_results <- c(r2 = summary(model)$r.squared, 
                        slope = unname(coef(model)[2]),
                        var_residuals = var(model$residuals))
print(regression_results)


  # extract the time series of the scalar projection of the optimum on the direction of optimum oscillation
shifted_optimum_proj <- shift_opt_proj(traj_chang_env, lag)

  # transform back to the optimum vector
shifted_optimum <- 
  t(sapply(shifted_optimum_proj, function(x) x*traj_chang_env$direction))
  
  # plot trajectory and shifted optimum
ggplot(data.frame(time = traj_chang_env$time,
                  x1 = traj_chang_env$traj[,1],
                  theta1 = shifted_optimum[,1]),
       aes(time)) +
  geom_line(aes(y = theta1, color = "theta1")) +
  geom_line(aes(y = x1, color = "x1")) +
  labs(x = "Time", y = expression(paste(x[1],"/",theta[1])),
       title = "Trajectory w/ shifted optimum",
       color = "") +
  scale_color_manual(values = c(x1 = "black", theta1 = "#E89143"),
                     labels = c(x1 = expression(x[1]), theta1 = expression(theta[1]))) +
  my_theme()

# 2.3 Spectrum analysis and peak identification
# Identifying the signal of environmental change from the spectral density of the phenotype time-series alone.

  # calculate the scalar projection of the phenotype on the direction of environmental change, and create the corresponding ts (time-series) object
pheno <- scalar_projection(traj_chang_env$traj, traj_chang_env$direction)
pheno_ts <- ts(pheno, frequency = 1/traj_chang_env$dt)

  # calculate spectral density of the phenotype time-series
traj_spec <- spectrum(pheno_ts, plot = F)

  # identify the peak of spectral dnesity and its correspondent period
  # (1/frequency)
peak <- c(peak_period = 1/traj_spec$freq[which.max(traj_spec$spec)],
          peak_spectral_dens = max(traj_spec$spec))

  # plot spectral density with indicated peak, and print peak information to the console
plot(1/traj_spec$freq, traj_spec$spec, type = "l",
     xlab = "Period",
     ylab = "Spectral density",
     main = "Trajectory spectral density")
points(peak[1], peak[2], pch = 16, col = "red")

print(peak)

# 2.4 Estimation of regression slope
# Estimate the amplitude of the tracking component of the trajectories from the spectral density of phenotype alone. 
# Estimate the regression slope previously calculated in section 2.2 as the ratio of this amplitude to that of optimum.
# This estimation is done for a series of interpolated time-series of phenotype, finally selecting the highest estimate.

  # First, generate list of interpolated phenotype time-series:
  # Each element of the list is a time-series incremented by one extra data-point (up to doubling the number of data points)
length_pheno <- length(pheno)
interp_phenos <- 
  sapply(0:(1*length_pheno), function(n){
    approx(pheno, n = length_pheno + n)$y
  })

  # estimate the slopes for each of the interpolated time-series
est_slopes <- sapply(seq_along(interp_phenos), function(i){
  
    # calculate the spectral density for the current interpolated time-series
  pheno_ts <- ts(interp_phenos[[i]], frequency = 1/traj_chang_env$dt)
  traj_spec <- spectrum(pheno_ts, plot = F)
    
    # identify peak of spectral density
  max_spec <- max(traj_spec$spec)
  freq <- traj_spec$freq[which.max(traj_spec$spec)]
    
    # Identify the resolution of frequency values for which the spectral density is measured by the fast-Fourier transform method used by spectrum()
    # This resolution depends on the number of data-points
  resolution <- tail(traj_spec$freq, 1)/length(traj_spec$freq)
  
    # Estimate slope of the tracking component and the according
    # Estimation  the regression slope
  est_amp <- 2*sqrt(max_spec*resolution)
  est_slope <- est_amp/5.99
  
  return(est_slope)
})

  # find position of the maximum estimated slope
max_slope_position <- which.max(est_slopes)

  # plot estimated slopes and identify the highest one
plot(est_slopes,
     xlab = "Number of points added to the phenotype time series",
     ylab = "Estimated slope",
     main = "Estimated slopes for the serially interpolated time series")
points(max_slope_position, est_slopes[max_slope_position], col = "red", pch = 16)

  # Print regression results again and currently estimated slope for comparison  
print(regression_results)
print(c(peak_estimated_slope = est_slopes[max_slope_position]))


# 3. MODEL WITHOUT INTERACTION -------------------------------------------------
# 3.1 Simulate a single trajectory for a given condition of optimum oscillation
d <- 50
amplitude <- 5.99
period <- 3

traj_wo_int <- gen_traj(d = d, tmax = 100, rescale = T,
                        coefs = c(0, 0, 1),
                        direction = rep(1/sqrt(d), d),
                        ch.opt.FUN = function(t) ch.opt.sinusoid(t, amplitude,  period = 3))

  # extract optimum for plotting
optimum <- get.opt(traj_wo_int)

ggplot(data.frame(time = traj_wo_int$time,
                  x1 = traj_wo_int$traj[,1],
                  theta1 = optimum[,1]),
       aes(time)) +
  geom_line(aes(y = theta1, color = "theta1")) +
  geom_line(aes(y = x1, color = "x1")) +
  labs(x = "Time", y = expression(paste(x[1],"/",theta[1])),
       title = "Trajectory w/o interactions",
       color = "") +
  scale_color_manual(values = c(x1 = "black", theta1 = "#E89143"),
                     labels = c(x1 = expression(x[1]), theta1 = expression(theta[1]))) +
  my_theme()

# 3.2 Calculate the gradient
# compute the absolute value of the directional selection gradient along the direction of change in the optimum
abs_gradient <- abs(scalar_projection(-(traj_wo_int$traj - optimum)^3, traj_wo_int$direction))

  # Print mean selection gradient
result <- c(mean_abs_gradient = mean(abs_gradient))
print(result)

  # Plot gradient together with trajectory
index_window <- 1:151
y_range <- range(c(traj_wo_int$traj[,1],
                   optimum[,1],
                   abs_gradient))
line_width <- 2
time_window <- range(traj_wo_int$time[index_window])

{
  op <- par(mar = c(4.5, 4.5, 1, 4.5))
  plot(traj_wo_int$time[index_window], traj_wo_int$traj[index_window, 1], 
       type = "l", ylim = y_range, lwd = line_width, 
       xlab = "Time", ylab = expression(paste(x[1], ',', theta[1])), 
       axes = F)
  lines(traj_wo_int$time[index_window], optimum[index_window, 1], 
        col = "#E89143", lwd = line_width)
  box(lwd = line_width)
  axis(1, lwd = line_width, las = 1, tck = -0.03, 
       at = seq(0, time_window[2], time_window[2]/4))
  axis(2, lwd = line_width, las = 1, tck = -0.03)
  par(new = T)
  plot(traj_wo_int$time[index_window], abs_gradient[index_window], 
       type = "l", ylim = y_range, col = "olivedrab3",
       ylab = "", lwd = line_width, xlab = "", 
       axes = F)
  axis(4, col = "olivedrab3", lwd = line_width, las = 1, tck = -0.03)
  mtext(side = 4, line = 3, "K(x)")
  par(op)
}