# draw.x() ---------------------------------------------------------------------

# DESCRIPTION
# This function is used to draw initial phenotypes for the simulations. 
# It draws n random vectors in the d-dimensional phase state and returns them in a matrix format with vectors organized by rows. 
# Vector are drawn from multivariate normal distributions, the standard deviation of which varies between methods. 
# Method doebeli2014 sets the standard deviation as d. With method average.k, the SD is chosen such that the expected K(x) is k.

# ARGUMENTS
# d             number of dimensions/traits;
# n             number of vectors to be drawn;
# method        sets the standard deviation of the distribution as described above (either "doebeli2014" or "average.k");
# k             Average K(x)  of the vectors, when using the "average.k" method;

draw.x <- function(d = 2, n = 1, method = "average.k", k = 0.5){
  
  if(method == "doebeli2014") sd <- d
  if(method == "average.k") sd <- (-4*log(k)/(3*d))^(1/4)
  
  state <- NULL
  for(i in 1:n)
  {
    cur.state <- rnorm(d, mean = 0, sd = sd)
    state <- rbind(state, cur.state)
  }
  row.names(state) <- NULL
  
  return(state)
  
}

# draw.kernel() ----------------------------------------------------------------

# DESCRIPTION
# The function creates a list containing two arrays of interaction coefficients corresponding to the aij and bij coefficients in equation (1) in the main text. 
# The bidimensional b array refers to first-order interactions and the three-dimensional a array, to the second order-interactions. 
# Values are drawn from a standard normal distribution.

# ARGUMENTS
# d                 is the number of traits;

draw.kernel <- function(d = 2) {
  
  b <- array(rnorm(d^2, mean = 0, sd = 1), dim = c(d, d))
  a <- array(rnorm(d^3, mean = 0, sd = 1), dim = c(d, d, d))
  
  kernel <- list("b" = b, "a" = a)
  
  return(invisible(kernel))
}

# gen_traj() -------------------------------------------------------------------
# calc_traj()

# DESCRIPTION
# These functions are used to numerically solve for the dynamics of the system by using the "lsoda" method implemented in the package deSolve. 
# The function calc_traj() performs the calculation and is called internally, while gen_traj() is called by the user. 
# Values for the initial condition vector and the kernel arrays can be set in the arguments, otherwise they will be drawn from draw.x and draw.kernel. 
# Standardization of the dynamics across dimensionalities can be applied by rescaling kernel coefficients b by sqrt(d) and a by d (as done in  all simulations used in the paper). 
# If the unstandardized model (rescale = FALSE), the timescale (interval for numerical integration and maximum time) is divided by d^2 (also acchieved by setting tmax.exact and dt.exact to FALSE)
# Different weights can be applied to each term in the RHS of equation (1) in the main text using coefs. In particular, freq-dep interactions are removed by setting coefs to c(0, 0, 1). 
# The regime of optimum oscillation is determined by a function ch.opt.FUN which should only take time as an argument. 


# ARGUMENTS
# d               number of dimensions/traits;
# xo              predetermined initial condition in the form of either a vector or a matrix of one row;
# kernel          kernel list (such as the output of draw.kernel) to be used for the approximation;
# coef            numeric vector of coefficients to be appied to the first, second and third terms of the ode, respectively
# start.time      starting time time for the numerical integration 
# tmax            total duration over which the dynamics are numerically integrated 
# tmax.exact = T  boolean variable determining whether the maximum time should be used unmodified, versus divided by d^2 
# dt              time step for numerical integration;
# dt.exact = T    boolean variable determining whether the time interval should be used unmodified, versus divided by d^2
# rescale         boolean variable determining whether rescaling must be applied to interaction kernel arrays
# ch.opt.FUN      function (of time only) describing the dynamics of the optimum; if NULL, optimum is constant at zero.

# OUTPUT
# The function returns a list:
# $traj           matrix containing the  trajectory
# $dt and $tmax   time interval and maximum duration of the numerical integration. These may be scaled by d^2 if the unstandardized interaction kernel has been used.
# $kernel         list with the arrays a and b of the unstandardized competition kernel; 
# $time           vector with the time values at each step, starting at start.time up to start.time+tmax, in steps of dt
# $rescaled       boolean indicating whether the dynamics was rescaled 
# $coef           vector of coefficients used in the arguments
# $ch.opt.FUN     function (of time only) describing the dynamics of the optimum
# $direction      direction of optimum oscillation in multivariate space


calc_traj <- function(xo, times, kernel_a, kernel_b, direction, coefs, ch.opt.FUN){
  params <- list(kernel_a = kernel_a, 
                 kernel_b = kernel_b, 
                 direction = direction,
                 coefs = coefs,
                 ch.opt.FUN = ch.opt.FUN)
  
  system <- function(t, x, params){
    with(params, {
      first <- kernel_b%*%x
      
      local.function <- function(a_slice){
        sum(a_slice*outer(x, x))
      }
      second <- apply(kernel_a, 1, local.function)
      
      theta <- direction*ch.opt.FUN(t)
      
      third <- -(x - theta)^3
      
      list(as.vector(coefs[1]*first + coefs[2]*second + coefs[3]*third))
    })
  }
  
  out <- ode(y = xo, 
             times = times, 
             func = system, 
             parms = params, 
             method = "lsoda")
  
  return(out[,-1])
}


gen_traj <- function(d = 2, xo = NULL, kernel = NULL, coefs = c(1, 1, 1), start.time = 0, tmax = 400, tmax.exact = F, dt = 0.1, dt.exact = F, rescale = T, ch.opt.FUN = NULL, direction = 1){
  
  require(deSolve)
  
  if(rescale == T) {
    tmax.exact = T
    dt.exact = T
  }
  
  if(tmax.exact == F) tmax <- tmax/d^2
  if(dt.exact == F) dt <- dt/d^2
  
  times <- seq(start.time, start.time+tmax, dt)
  
  if(is.null(xo)) {
    xo <- draw.x(d)
  }
  xo <- as.vector(xo)
  
  if(is.null(kernel)) kernel <- draw.kernel(d)
  
  if(rescale)
  {
    kernel_b <- kernel$b/sqrt(d)
    kernel_a <- kernel$a/d
  }
  
  if(is.null(ch.opt.FUN))
    eval(parse(text = sprintf("ch.opt.FUN <- function(t) rep(0, %s)", d)))
  
  if(body(ch.opt.FUN)[[1]] == "ch.opt.sinusoid"){
    body(ch.opt.FUN)[[3]] <- eval(body(ch.opt.FUN)[[3]])
    body(ch.opt.FUN)[[4]] <- eval(body(ch.opt.FUN)[[4]])
  }
  
  
  if(length(direction)==1){
    trait_n <- direction
    direction <- rep(0, d)
    direction[trait_n] <- 1
  } else {
    if(length(direction)!=d) stop("direction of wrong length")
    direction <- direction/vector_norm(direction)
  }
  
  traj <- calc_traj(xo, times, kernel_a, kernel_b, direction, coefs, ch.opt.FUN)
  
  out <- list("traj" = traj, 
              "time" = times,
              "kernel" = kernel,
              "dt" = dt,
              "tmax" = tmax,
              "rescaled" = rescale,
              "coefs" = coefs,
              "ch.opt.FUN" = ch.opt.FUN,
              "direction" = direction)
  
  environment(out$ch.opt.FUN) <- globalenv()
  
  return(invisible(out))
  
}


# vary.x() ---------------------------------------------------------------------
# DESCRIPTION
# This function is called by calc_lle() to produce a small phenotypic deviation with direction defined by a unitary vector, and a given magnitude. 
# If no direction is set, a random direction is picked. 

# ARGUMENTS
# state       mutlivariate phenotype from which a deviation is picked
# do          magnitude of the deviation (norm of the vector)
# direction   unitary vector giving the direction of the deviation 

vary.x <- function(state, do = 1e-8, direction = NULL, check_direction = T) {
  
  d <- length(state)
  
  if(check_direction){
    if(is.null(direction)){
      direction <- rnorm(d, 0, 1)
      direction <- direction/vector_norm(direction)
    } else {
      vec_norm <- vector_norm(direction)
      if((vec_norm - 1) > 1e-7){
        direction <- direction/vector_norm(direction)
        warning("direction was coerced to a unitary vector")
      }
    }
  }
  
  out <- state + direction*do
  
  return(out)
}

# calc_lle() -------------------------------------------------------------------
# DESCRIPTION
# This function calculates the Largest Lyapunov Exponent (LLE) for a given trajectory. 
# The calculation is done as described by Ispolatov (2015, Scientific Reports 5) and Spprot (http://sprott.physics.wisc.edu/chaos/lyapexp.htm). 
# An appropriate dt is <= 0.1/d^2 for unscaled trajectories or <= 0.1 for rescaled trajectories.
# If the trajectory does not contain elements $ch.opt.FUN or $direction, they will be added as NULL, function(state, theta) -(state^3) and c(1, 0, 0, ...), respectively.

# ARGUMENTS
# x             a trajectory that is an output of the function gen_traj();
# tau           exact time the deviated state is allowed diverge before calculation of the LE;
# do            inital distance between base and deviated trajectories;

# OUTPUT
# A list of two vectors elements: 
# the calculated LLE along the trajectory ($lle); 
# the time points at which each iteration of the calculation of the LLE begins in the correspondent trajectory ($time).

calc_lle <- function(x, tau = 0.1, do = 1e-3, ...){

  if(tau%%x$dt!=0) 
    stop("the provided tau was not a multiple of the dt of the provided trajectory")

  d <- ncol(x$traj)
  
  it.start <- head(seq(1, nrow(x$traj), tau/x$dt), -1)
  values.lle <- vector(length = length(it.start))
  direction <- NULL
  if(x$rescaled){
    kernel_b <- x$kernel$b/sqrt(d)
    kernel_a <- x$kernel$a/d
  }
  dxts <- as.data.frame(matrix(ncol = d, nrow = length(it.start)))
  for(i in 1:length(it.start))
  {
    dxo <- vary.x(x$traj[it.start[i],], do = do, direction = direction)
    dxt <- tail(calc_traj(xo = dxo, 
                          times = seq(x$time[it.start[i]], x$time[it.start[i]] + tau, x$dt), 
                          kernel_a = kernel_a, 
                          kernel_b = kernel_b,
                          direction = x$direction, 
                          coefs = x$coefs, 
                          ch.opt.FUN = x$ch.opt.FUN), 1)
    dxt <- as.vector(dxt)
    
    dxts[i,] <- dxt
    
    sep_vec <- dxt - x$traj[it.start[i]+tau/x$dt,]
    
    distance <- vector_norm(sep_vec)
    
    cur.le <- log(distance/do)/tau
    
    values.lle[i] <- cur.le
    
    direction <- sep_vec/distance
  }
  
  time <- x$time[it.start]
  
  out <- list(lle = values.lle,
              time = time)
  
  return(out)
}

# calc_local_lle() -------------------------------------------------------------
# DESCRIPTION
# This function calculates the window average of the LLE from an output of calc_lle(), as described in the SI. 
# A burnout is included to discard the beginning of the LLE time-series.

# ARGUMENTS
# lle               the output of calc_lle() from which the average will be calculated
# average_window    window size in units of time 
# burnout           initial period of time for which the LLE time-series will be discarded

# OUTPUT
# A list with:
# $local_lle        time-series of the window average of the LLE
# $time             vector of time points

calc_local_lle <- function(lle, average_window = 200, burnout = 50){
  dt <- diff(lle$time[1:2])
  window_n_steps <- average_window/dt
  burnout_n_steps <- burnout/dt
  
  if(window_n_steps + burnout_n_steps > length(lle$lle))
    stop("ERROR: trajectory is shorter than average_window + burnout.")
  
  index <- (window_n_steps + burnout_n_steps + 1):length(lle$lle)
  
  local_lle <- sapply(index, function(i) mean(lle$lle[(i - window_n_steps):i]))
  
  out <- list(local_lle = local_lle,
              time = lle$time[index])
  
  return(out)
}

# class_lle() ------------------------------------------------------------------
# DESCRIPTION
# This function classifies a trajectory along time according to its local LLE into equilibrium, cyclic or chaotic.
# A magnitude parameter (criterion below) is used to assess whether LLE is substantially larger/smaller than zero. 

# ARGUMENTS
# local_lle           the output of calc_local_lle() which will be classified;
# criterion           classification criterion (greek letter mu in the SI)

# OUTPUT
# A list: 
# $class              classification of the trajectory at each time point
# $time               vector of time points
class_lle <- function(local_lle, criterion = 0.02){
  if(criterion < 0){
    criterion <- abs(criterion)
    warning("criterion coerced to abs(criterion)")
  }
  
  class <- sapply(local_lle$local_lle, function(x) {
    if(x > criterion){
      return(3)
    } else if(x < -criterion){
      return(1)
    } else 
      return(2)
  })
  
  class <- factor(class, levels = c(1, 2, 3), labels = c("equilibrium", 
                                                         "cyclic", 
                                                         "chaos"))
  
  out <- list(class = class,
              time = local_lle$time)
  
  return(out)
}

# ch.opt.sinusoid() ------------------------------------------------------------
# DESCRIPTION
# This function is passed to gen_traj() to describe the optimum oscillation regime. 
# It produces a sine function in time t, with given amplitude and period.

# ARGUMENTS
# t                   time
# amplitude           amplitude of oscillation;
# period              period of oscillation;

ch.opt.sinusoid <- function(t, amplitude, period){
  amplitude*sin(2*pi*t/period)
}

# get.opt() --------------------------------------------------------------------
# DESCRIPTION
# This function returns the optimum of a given trajectory (object created with gen_traj()). 
# The main purpose of this function is to allowing plotting the optimum of a trajectory

# ARGUMENTS
# x       trajectory (output object from gen_traj());

# OUTPUT
# A matrix where each row is a vector of optimum phenotypes for d traits, with as many rows as time points

get.opt <- function(x){
  out <- sapply(x$time, function(t) (x$ch.opt.FUN(t) * x$direction))
  if(is.matrix(out)) out <- t(out)
  return(out)
}

# scalar_projection() ----------------------------------------------------------
# DESCRIPTION
# Function for calculating the scalar projection of vectors in the direction of another vector. 
# Direction vector do not need to be passed as unitary vectors, as it will be coerced into it by the function. 

# ARGUMENTS
# vec           single vector or matrix of vectors to be projected
# direction     single vector indicating the direction towards which vec should be projected

scalar_projection <- function(vec, direction){
  direction <- direction/vector_norm(direction)
  return(vec%*%direction)
}

# vector_norm() ----------------------------------------------------------
# DESCRIPTION
# Calculates the norm (length) of a single vector.

# ARGUMENTS
# vec       vector 


vector_norm <- function(vec){
  return(sqrt(sum(vec^2)))
}

# shift_opt_proj() ----------------------------------------------------------
# best_shift()
# regression_model()

# DESCRIPTION
# Functions used for correcting the lag between phenotype and optimum oscillations in trajectories under oscilalting optimum conditions. 
# shift_opt_proj() generates a shifted time-series of the scalar projection of the optimum in the direction of optimum oscillation. 
# This is used by best_shift(), which returns the shift that maximizes the fit of the linear regression of the scalar projection of phenotype on the scalar projection of the optimum. 
# The maximum shift explored by the optimization process is half a period of optimum oscillation.
# regression_model() is a wrapper that returns the regression model between phenotype and optimum scalar projections for a given trajectory (output of gen_traj()) and optimum shift.

# ARGUMENTS
# traj       trajectory (output of gen_traj()) on which the operations are applied
# shift      shift (delay in units of time) of the optimum time-series
# period     period of optimum oscillation used in the simulation

shift_opt_proj <- function(traj, shift){
  out <- sapply(traj$time, function(t) traj$ch.opt.FUN(t - shift))
  return(out)
}

best_shift <- function(traj, period){
  pheno <- scalar_projection(traj$traj, traj$direction)
  
  out <- optim(0, function(shift) {
    opt <- shift_opt_proj(traj, shift)
    
    model <- lm(pheno~opt)
    
    if(model$coefficients[2] < 0) # discard fits of negative slope
      out <- 0
    else
      out <- summary(model)$r.squared
    
    return(1 - out)
  }, 
  lower = 0, upper = 0.5*period, # set the upper bound of the shift as half the
                                 # period of optimum oscillation
  method = "Brent") # method for one-dimensional maximization
  
  out <- out$par
  
  return(out)
}

regression_model <- function(traj, shift = 0){
  pheno <- scalar_projection(traj$traj, traj$direction)
  opt <- shift_opt_proj(traj, shift)
  
  model <- lm(pheno~opt)
  
  return(model)
}


# my_theme() -------------------------------------------------------------------
# DESCRIPTION
# Ggplot2 theme used across some of the plots in example.R.

my_theme <- function(){
  theme(panel.background = element_rect(color = 1, fill = "white", size = 1),
        panel.grid.major = element_blank(),
        panel.grid.minor = element_blank(),
        text = element_text(size = 10, 
                            color = "black"),
        axis.text = element_text(color = "black"),
        legend.key = element_rect(fill = "white"),
        legend.background = element_rect(fill = "transparent"))
}

# load_install_pachage() -------------------------------------------------------
# DESCRIPTION
# Verifies whether a given R package is installed in the machine. If it is, the package is loaded. If not, the package is installed and then loaded.

# ARGUMENTS
# pckg        string corresponding to package name

load_install_package <- function(pckg){
  if(!require(pckg, character.only = T)){
    install.packages(pckg)
    library(pckg, character.only = T)
  }
}